import sys
import os
from pycocotools import mask
import numpy as np
import cv2
import json
import functools
from collections import defaultdict
import torch
from tqdm import tqdm
from torch.utils.data import Dataset
import torchvision
from torchvision import datasets, transforms
from torch.utils.data import DataLoader

from pymlutil.s3 import s3store, Connect
from pymlutil.jsonutil import ReadDict
from pymlutil.imutil import ImUtil, ImTransform, AddGaussianNoise, ResizePad

default_loaders = [{'set':'train', 'enable_transform':True, 'shuffle': True},
                   {'set':'test', 'enable_transform':False, 'shuffle': False}]

def CreateCifar10Loaders(dataset_path, batch_size = 2,  
                      num_workers=0, cuda = True, loaders = default_loaders, 
                      rotate=3, scale_min=0.75, scale_max=1.25, offset=0.1, augment_noise=0.0, width=32, height=32):

    Cifar10Classes = ('plane', 'car', 'bird', 'cat', 'deer', 'dog', 'frog', 'horse', 'ship', 'truck')

    pin_memory = False
    # if cuda:
    #     pin_memory = True

    for i, loader in enumerate(loaders):
        if loader['enable_transform']:
            transform_list = []

            if width != 32 or height != 32:
                transform_list.append(ResizePad(width, height))

            transform_list.append(transforms.RandomHorizontalFlip(p=0.5))
            if rotate > 0 or offset > 0 or scale_min != 1.0 or scale_max != 1.0:
                transform_list.append(transforms.RandomAffine(degrees=rotate,
                        translate=(offset, offset), 
                        scale=(scale_min, scale_max), 
                        interpolation=transforms.InterpolationMode.BILINEAR))
            transform_list.append(transforms.ToTensor())
            transform_list.append(transforms.Normalize((0.0, 0.0, 0.0), (1.0, 1.0, 1.0))) # Imagenet mean and standard deviation
            if augment_noise > 0.0:
                transform_list.append(AddGaussianNoise(0., augment_noise))

            transform = transforms.Compose(transform_list)
        else:
            transform_list = []
            if width != 32 or height != 32:
                transform_list.append(ResizePad(width, height))
            transform_list.append(transforms.ToTensor())
            transform_list.append(transforms.Normalize((0.0, 0.0, 0.0), (1.0, 1.0, 1.0))) # Imagenet mean and standard deviation

            transform = transforms.Compose(transform_list)

        dataset = torchvision.datasets.CIFAR10(root=dataset_path, train=loader['set']=='train', download=True, transform=transform)
        # Creating PT data samplers and loaders:
        loader['width']=width
        loader['height']=height
        loader['in_channels']=3
        loader['num_classes']=len(Cifar10Classes)
        loader['classes']=list(Cifar10Classes)
        loader['batches'] =int(len(dataset)/batch_size)
        loader['length'] = loader['batches']*batch_size
        loader['dataloader'] = torch.utils.data.DataLoader(dataset=dataset, 
                                                    batch_size=batch_size,
                                                    shuffle=loader['shuffle'],
                                                    num_workers=num_workers, 
                                                    pin_memory=pin_memory)
    return loaders

def main(args):

    parameters = ReadDict(args.test_config)

    if args.test_dataset:
        loaders = CreateCifar10Loaders(args.dataset_path, 
                        batch_size = args.batch_size,  
                        num_workers=args.num_workers, 
                        cuda = args.cuda, 
                        rotate=args.augment_rotation, 
                        scale_min=args.augment_scale_min, 
                        scale_max=args.augment_scale_max, 
                        offset=args.augment_translate_x,
                        augment_noise = args.augment_noise,
                        width = parameters['cifar10']['width'],
                        height = parameters['cifar10']['height'],
                        )

        parameters['cifar10']['test_path']=os.path.join(parameters['cifar10']['test_path'], '') # Add trailing slash if not present
        os.makedirs(parameters['cifar10']['test_path'], exist_ok=True)

        for loader in tqdm(loaders, desc="Loader"):
            for i, data in tqdm(enumerate(loader['dataloader']), 
                                desc="Batch Reads", 
                                total=loader['batches']):
                sample, target = data
                assert(sample is not None)
                assert(sample.shape[0] == args.batch_size)
                assert(sample.shape[2] == loader['height'])
                assert(sample.shape[3] == loader['width'])
                assert(target is not None)

                sample =  sample.permute(0, 2, 3, 1) # Change to batch, height, width, channel for rendering
                sample_max = sample.max()
                sample_min = sample.min()
                if sample_max > sample_min:
                    for j, image in enumerate(sample):
                        image = 255*(image - sample_min)/(sample_max-sample_min) # Convert to RGB color rane
                        image = image.cpu().numpy().astype(np.uint8)
                        image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
                        write_path = '{}{}{:03d}{:03d}.png'.format(parameters['cifar10']['test_path'], loader['set'], i,j)            
                        cv2.imwrite(write_path,image)

                if args.num_images is not None and args.num_images > 0 and i >= args.num_images:
                    print ('test_dataset complete')
                    break

    print('Test complete')

#objdict = json.load(open('/data/git/mllib/datasets/coco.json'))
#Test(objdict, '/store/Datasets/coco/instances_val2017.json', '/store/Datasets/coco/val2014', 'COCO_val2014_')

def parse_arguments():
    parser = argparse.ArgumentParser(description='Process arguments')

    parser.add_argument('-d', '--debug', action='store_true',help='Wait for debuggee attach')   
    parser.add_argument('-debug_port', type=int, default=3000, help='Debug port')
    parser.add_argument('-debug_address', type=str, default='0.0.0.0', help='Debug port')
    parser.add_argument('-credentails', type=str, default='creds.yaml', help='Credentials file.')
    parser.add_argument('-dataset_path', type=str, default='./dataset', help='Local dataset path')
    parser.add_argument('-num_images', type=int, default=10, help='Number of images to display')
    parser.add_argument('-num_workers', type=int, default=0, help='Data loader workers')
    parser.add_argument('-batch_size', type=int, default=4, help='Dataset batch size')
    parser.add_argument('-ds', action='store_true', help='True to test dataset')
    parser.add_argument('-test_dataset', action='store_true', help='True to test dataset')
    parser.add_argument('-test_path', type=str, default='./datasets_test/', help='Test path ending in a forward slash')
    parser.add_argument('-test_config', type=str, default='test.yaml', help='Test configuration file')

    parser.add_argument('-cuda', type=bool, default=True, help='pytorch CUDA flag') 
    parser.add_argument('-numTries', type=int, default=3, help="Read retries")

    parser.add_argument('-augment_rotation', type=float, default=0.0, help='Input augmentation rotation degrees')
    parser.add_argument('-augment_scale_min', type=float, default=1.00, help='Input augmentation scale')
    parser.add_argument('-augment_scale_max', type=float, default=1.00, help='Input augmentation scale')
    parser.add_argument('-augment_translate_x', type=float, default=0.0, help='Input augmentation translation')
    parser.add_argument('-augment_translate_y', type=float, default=0.0, help='Input augmentation translation')
    parser.add_argument('-augment_noise', type=float, default=0.1, help='Augment image noise')

    args = parser.parse_args()

    if args.ds:
        args.test_dataset = True

    return args
    
if __name__ == '__main__':
    import argparse
    args = parse_arguments()

    if args.debug:
        print("Wait for debugger attach on {}:{}".format(args.debug_address, args.debug_port))
        import debugpy

        debugpy.listen(address=(args.debug_address, args.debug_port)) # Pause the program until a remote debugger is attached
        debugpy.wait_for_client()  # Pause the program until a remote debugger is attached
        print("Debugger attached")

    main(args)




from dataclasses import replace
import sys
import os
from pycocotools import mask
import numpy as np
import cv2
import json
from collections import defaultdict
import unittest
import torch
from tqdm import tqdm
from torch.utils.data import Dataset
from torchvision import datasets, transforms
from torch.utils.data import DataLoader, WeightedRandomSampler

from pymlutil.s3 import s3store, Connect
from pymlutil.jsonutil import ReadDict
from pymlutil.imutil import ImUtil, ImTransform

sys.path.insert(0, os.path.abspath('')) # Test files from current path rather than installed module
from torchdatasetutil.cityscapesstore import CreateCityscapesLoaders

test_config = 'test.yaml'
class Test(unittest.TestCase):      

    
    def test_CreateImageLoaders(self):
        parameters = ReadDict(test_config)

        if 'images' not in parameters:
            raise ValueError('images not in {}'.format(test_config))

        s3, creds, s3def = Connect(parameters['images']['credentials'])

        class_dictionary = ReadDict(parameters['cityscapes']['class_dict'])
        imUtil = ImUtil({}, class_dictionary)

        sampler =parameters['cityscapes']['sampler']
        if sampler and 'sample_weights' in class_dictionary:
                train_sampler_weights = class_dictionary['sample_weights']['weights']
                upsampled_class = class_dictionary['sample_weights']['class']
        else:
            train_sampler_weights=None
            upsampled_class=None



        loaders = CreateCityscapesLoaders(s3, s3def, 
                        src = parameters['cityscapes']['obj_src'],
                        dest = parameters['cityscapes']['destination'],
                        class_dictionary =class_dictionary,
                        bucket = s3def['sets']['dataset']['bucket'], 
                        width=parameters['cityscapes']['width'], 
                        height=parameters['cityscapes']['height'], 
                        batch_size = parameters['cityscapes']['batch_size'], 
                        num_workers=parameters['cityscapes']['num_workers'],
                        train_sampler_weights=train_sampler_weights,)

        parameters['cityscapes']['test_path']=os.path.join(parameters['cityscapes']['test_path'], '') # Add trailing slash if not present
        os.makedirs(parameters['cityscapes']['test_path'], exist_ok=True)

        minority_class_list = []

        for loader in tqdm(loaders, desc="CreateImageLoaders"):
            for i, data in tqdm(enumerate(loader['dataloader']), 
                                desc="Batch Reads", 
                                total=loader['batches']):
                inputs, labels, mean, stdev = data
                assert(inputs.size(0)==parameters['cityscapes']['batch_size'])
                assert(inputs.size(-1)==parameters['cityscapes']['width'])
                assert(inputs.size(-2)==parameters['cityscapes']['height'])
                assert(labels.size(0)==parameters['cityscapes']['batch_size'])
                assert(labels.size(-1)==parameters['cityscapes']['width'])
                assert(labels.size(-2)==parameters['cityscapes']['height'])

                images = inputs.cpu().permute(0, 2, 3, 1).numpy()
                labels = np.around(labels.cpu().numpy()).astype('uint8')
                mean = mean.cpu().numpy()
                stdev = stdev.cpu().numpy()


                for j, image in enumerate(images):
                    if upsampled_class is not None:
                        minority_class_list.append(upsampled_class in labels[j])

                    img = imUtil.MergeIman(images[j], labels[j], mean[j], stdev[j])
                    write_path = '{}{}{:03d}{:03d}.png'.format(parameters['cityscapes']['test_path'], loader['set'], i,j)                   
                    cv2.imwrite(write_path,img)

                #    is_success, buffer = cv2.imencode(".png", img)
                #    if not is_success:
                #        raise ValueError('test_imstore test_CreateImageLoaders cv2.imencode failure batch {} image {}'.format(i, j))

                if sampler and i >= 200:
                    minority_class_ratio = sum(minority_class_list)/len(minority_class_list)
                    if minority_class_ratio < 0.2:
                        raise ValueError('Weighted Random Sampler maybe dysfunctional. {:4f} ratio is too low for specified minority class'.format(minority_class_ratio))
                    break

                if 'test_images' in parameters['cityscapes'] and parameters['cityscapes']['test_images'] is not None and i*parameters['cityscapes']['batch_size'] >= parameters['cityscapes']['test_images']:
                    break
            
            sampler=False

if __name__ == '__main__':
    unittest.main()


import sys
import os
from pycocotools import mask
import numpy as np
import cv2
import json
from collections import defaultdict
import unittest
import torch
from tqdm import tqdm
from torch.utils.data import Dataset
from torchvision import datasets, transforms
from torch.utils.data import DataLoader

from pymlutil.s3 import s3store, Connect
from pymlutil.jsonutil import ReadDict
from pymlutil.imutil import ImUtil, ImTransform

# Force import from local files
sys.path.insert(0, os.path.abspath('')) # Test files from current path rather than installed module
from torchdatasetutil.cocostore import CocoStore, CreateCocoLoaders

test_config = 'test.yaml'
class Test(unittest.TestCase):      

    def test_iterator(self):
        parameters = ReadDict(test_config)

        if 'coco' not in parameters:
            raise ValueError('coco not in {}'.format(test_config))

        s3, creds, s3def = Connect(parameters['coco']['credentials'])

        dataset_desc = s3.GetDict(s3def['sets']['dataset']['bucket'],parameters['coco']['dataset_val'])
        class_dictionary = s3.GetDict(s3def['sets']['dataset']['bucket'],parameters['coco']['class_dict']) 
        imUtil = ImUtil(dataset_desc, class_dictionary)
                
        store = CocoStore(s3, bucket=s3def['sets']['dataset']['bucket'], 
                        dataset_desc=parameters['coco']['dataset_train'], 
                        image_paths=parameters['coco']['train_image_path'], 
                        class_dictionary=parameters['coco']['class_dict'])


        for i, iman in  tqdm(enumerate(store),
                             desc="CocoStore Reads",
                             total=len(store)):

            #img = store.MergeIman(iman['img'], iman['ann'])
            assert(iman['img'] is not None)
            assert(iman['ann'] is not None)
            if 'test_images' in parameters['coco'] and parameters['coco']['test_images'] is not None and i >= parameters['coco']['test_images']:
                break


    def test_CreateCocoLoaders(self):
        parameters = ReadDict(test_config)

        if 'coco' not in parameters:
            raise ValueError('coco not in {}'.format(test_config))

        s3, creds, s3def = Connect(parameters['coco']['credentials'])

        dataset_desc = s3.GetDict(s3def['sets']['dataset']['bucket'],parameters['coco']['dataset_val'])
        class_dictionary = s3.GetDict(s3def['sets']['dataset']['bucket'],parameters['coco']['class_dict']) 
        imUtil = ImUtil(dataset_desc, class_dictionary)

        loaders = CreateCocoLoaders(s3, bucket=s3def['sets']['dataset']['bucket'],
                                     class_dict=parameters['coco']['class_dict'], 
                                     batch_size=parameters['coco']['batch_size'], 
                                     height=parameters['coco']['height'], 
                                     width=parameters['coco']['width'],
                                     num_workers=parameters['coco']['num_workers'])

        for loader in tqdm(loaders, desc="CreateCocoLoaders"):
            for i, data in tqdm(enumerate(loader['dataloader']), 
                                desc="Batch Reads", 
                                total=loader['batches']):
                inputs, labels, mean, stdev = data
                assert(inputs.size(0)==parameters['coco']['batch_size'])
                assert(inputs.size(-1)==parameters['coco']['width'])
                assert(inputs.size(-2)==parameters['coco']['height'])
                assert(labels.size(0)==parameters['coco']['batch_size'])
                assert(labels.size(-1)==parameters['coco']['width'])
                assert(labels.size(-2)==parameters['coco']['height'])
                
                if 'test_images' in parameters['coco'] and parameters['coco']['test_images'] is not None and i >= parameters['coco']['test_images']:
                    break


if __name__ == '__main__':
    unittest.main()


import sys
import os
from pycocotools import mask
import numpy as np
import cv2
import json
import functools
import random
from collections import defaultdict
import torch
from tqdm import tqdm
from torch.utils.data import Dataset
from torchvision import datasets, transforms
from torch.utils.data import DataLoader

from pymlutil.s3 import s3store, Connect
from pymlutil.jsonutil import ReadDict
from pymlutil.imutil import ImUtil, ImTransform

class CocoStore(ImUtil):

    def __init__(self, s3, # pymlutil.s3 s3 class object
                 bucket, # bucket name string
                 dataset_desc, # object in bucket containing coco format dataset definition 
                 image_paths, # path in bucket to dataset images
                 class_dictionary, # json or yaml class dictionary object described in  https://github.com/bhlarson/torchdatasetutil/blob/main/torchdatasetutil.ipynb#ClassDictionary
                 imflags=cv2.IMREAD_COLOR, # image colorspace to load in cv2 format
                 name_decoration='',  # Additional test to append to the filename to load
                 numTries=3 ): # Number of read retries

        self.s3 = s3 
        self.bucket = bucket 
        self.dataset_desc = s3.GetDict(bucket,dataset_desc) 
        self.class_dictionary = s3.GetDict(bucket,class_dictionary)
        self.image_paths = image_paths
        self.name_decoration = name_decoration
        self.imflags = imflags
        self.numTries = numTries



        self.CreateIndex()
        super(CocoStore, self).__init__(dataset_desc=self.dataset_desc, class_dictionary=self.class_dictionary)
        self.i = 0

    def CreateIndex(self):
        # create index objDict rather than coco types
        anns, cats, imgs = {}, {}, {}
        imgToAnns = defaultdict(list)
        catToImgs = defaultdict(list)
        catToObj = {}
        objs = {}
        # make list based on 
        if 'annotations' in self.dataset_desc:
            for ann in self.dataset_desc['annotations']:
                imgToAnns[ann['image_id']].append(ann)
                anns[ann['id']] = ann

        if 'images' in self.dataset_desc:
            for img in self.dataset_desc['images']:
                imgs[img['id']] = img

        if 'categories' in self.dataset_desc:
            for cat in self.dataset_desc['categories']:
                cats[cat['id']] = cat

        if 'annotations' in self.dataset_desc and 'categories' in self.dataset_desc:
            for ann in self.dataset_desc['annotations']:
                catToImgs[ann['category_id']].append(ann['image_id'])

        if self.class_dictionary  is not None:
            for obj in self.class_dictionary ['objects']:
                catToObj[obj['id']] = obj
                objs[obj['trainId']] = {'category':obj['category'], 
                                        'color':obj['color'],
                                        'display':obj['display']
                                        }


        # create class members
        self.anns = anns
        self.imgToAnns = imgToAnns
        self.catToImgs = catToImgs
        self.imgs = imgs
        self.cats = cats
        self.catToObj = catToObj
        self.objs = objs


    def __iter__(self):
        self.i = 0
        return self

    def classes(self, anns):
        class_vector = np.zeros(self.class_dictionary ['classes'], dtype=np.float32)

        for ann in anns:
            obj = self.catToObj[ann['category_id']]
            if obj['trainId'] < self.class_dictionary ["classes"]:
                class_vector[obj['trainId']] = 1.0

        return class_vector

    def DecodeImage(self, bucket, objectname):
        img = None
        
        for i in range(self.numTries):
            imgbuff = self.s3.GetObject(bucket, objectname)
            if imgbuff:
                imgbuff = np.frombuffer(imgbuff, dtype='uint8')
                try:
                    img = cv2.imdecode(imgbuff, flags=self.imflags)
                except:
                    print ("CocoStore::DecodeImage {}/{} cv2.imdecode exception i={}".format(bucket, objectname, i))
                    img = None
            if img is None:
                print('CocoStore::DecodeImage failed to load {}/{} try {}'.format(bucket, objectname, i))
            else:
                break
        return img

    def drawann(self, imgDef, anns):
        annimg = np.zeros(shape=[imgDef['height'], imgDef['width']], dtype=np.uint8)
        for ann in anns:
            obj = self.catToObj[ann['category_id']]
            if obj['trainId'] < self.class_dictionary ["classes"]:
                if type(ann['segmentation']) is list:
                    for i in range(len(ann['segmentation'])):
                        contour = np.rint(np.reshape(ann['segmentation'][i], (-1, 2))).astype(np.int32)
                        cv2.drawContours(image=annimg, contours=[contour], contourIdx=-1, color=obj['trainId'] , thickness=cv2.FILLED)
                elif type(ann['segmentation']) is dict:
                    rle = ann['segmentation']
                    compressed_rle = mask.frPyObjects(rle, rle.get('size')[0], rle.get('size')[1])
                    annmask = mask.decode(compressed_rle)
                    annimg[annmask] = obj['trainId']
                else:
                    print('unexpected segmentation')
            else:
                print('trainId {} >= classes {}'.format(obj['trainId'], self.class_dictionary ["classes"]))    
        return annimg

    def __len__(self):
        return len(self.dataset_desc['images'])

    def __next__(self):
        if self.i < self.__len__():
            result = self.__getitem__(self.i)
            self.i += 1
            return result
        else:
            raise StopIteration

    def __getitem__(self, idx):
        if idx >= 0 and idx < self.__len__():
            img_entry = self.dataset_desc['images'][idx]
            imgFile = '{}/{}{}'.format(self.image_paths,self.name_decoration,img_entry['file_name'])
            img = self.DecodeImage(self.bucket, imgFile)
            ann_entry = self.imgToAnns[img_entry['id']]
            ann = self.drawann(img_entry, ann_entry)
            if img is not None and ann is not None:
                classes = self.classes(ann_entry)
                result = {'img':img, 'ann':ann, 'classes':classes}
            else:
                result = None

            return result
        else:
            print('CocoStore.__getitem__ idx {} invalid.  Must be >=0 and < CocoStore.len={}'.format(idx, self.__len__()))
            return None

class CocoDataset(Dataset):
    def __init__(self, s3, bucket, dataset_desc, image_paths, class_dictionary, 
        height=640, 
        width=640, 
        imflags=cv2.IMREAD_COLOR, 
        image_transform=None,
        label_transform=None,
        name_decoration='',
        normalize=True, 
        enable_transform=True, 
        flipX=True, 
        flipY=False, 
        rotate=15, 
        scale_min=0.75, 
        scale_max=1.25, 
        offset=0.1,
        numTries=3, # Number of read retries
    ):
        self.image_transform = image_transform
        self.label_transform = label_transform
        self.height = height
        self.width = width
        self.imflags = imflags

        self.normalize = normalize
        self.enable_transform = enable_transform
        self.flipX = flipX
        self.flipY = flipY
        self.rotate = rotate
        self.scale_min = scale_min
        self.scale_max = scale_max
        self.offset = offset
        self.numTries = numTries

        self.store = CocoStore(s3, bucket, dataset_desc, image_paths, class_dictionary, imflags=self.imflags, name_decoration=name_decoration, numTries=self.numTries)


        self.imTransform = ImTransform(height=height, width=width, 
                                     normalize=normalize, 
                                     enable_transform=enable_transform, 
                                     flipX=flipX, flipY=flipY, 
                                     rotate=rotate, 
                                     scale_min=scale_min, scale_max=scale_max, offset=offset, astype=self.store.class_dictionary['input_type'])

    def __len__(self):
        return self.store.__len__()

    def __getitem__(self, idx):
        result = self.store.__getitem__(idx)
        if result is not None and result['img'] is not None and result['ann'] is not None:
            image = result['img']
            label = result['ann']

            if image is not None and label is not None:

                image, label, imgMean, imgStd = self.imTransform.random_resize_crop_or_pad(image, label)

                if len(image.shape) < 3:
                    image = np.expand_dims(image, axis=-1)

                image = torch.from_numpy(image).permute(2, 0, 1)
                label = torch.from_numpy(label)

                if self.image_transform:
                    image = self.image_transform(image)
                if self.label_transform:
                    label = self.label_transform(label)

                assert(image.shape[-1] == self.width)
                assert(image.shape[-2] == self.height)
            
        else:
            print('CocoDataset.__getitem__ idx {} returned result=None.'.format(idx))
            return None
        return image, label, imgMean, imgStd

# Handle corrupt images:
# https://github.com/pytorch/pytorch/issues/1137
# https://stackoverflow.com/questions/57815001/pytorch-collate-fn-reject-sample-and-yield-another/67583699#67583699
def collate_fn_replace_corrupted(batch, dataset, batch_size):
    original_batch_len = len(batch)
    batch = list(filter(lambda x: x is not None, batch)) # Filter out bad samples
    filtered_batch_len = len(batch)
    diff = original_batch_len - filtered_batch_len
    if filtered_batch_len < batch_size:                
        # Replace corrupted examples with another examples randomly
        print('imsgtore collate_fn_replace_corrupted diff = {}'.format(diff))
        batch.extend([dataset[random.randint(0, len(dataset))] for _ in range(diff)])
        # Recursive call to replace the replacements if they are corrupted
        return collate_fn_replace_corrupted(batch, dataset, batch_size)

    # Finally, when the whole batch is fine, return it
    return torch.utils.data.dataloader.default_collate(batch)

default_loaders = [{'set':'train', 'dataset': 'data/coco/annotations/instances_train2017.json', 'image_path':'data/coco/train2017' , 'enable_transform':True},
                   {'set':'test', 'dataset': 'data/coco/annotations/instances_val2017.json', 'image_path':'data/coco/val2017', 'enable_transform':False}]

def CreateCocoLoaders(s3, bucket, class_dict, 
                      batch_size = 2, shuffle=True, 
                      num_workers=0, cuda = True, timeout=0, loaders = default_loaders, 
                      height=640, width=640, 
                      image_transform=None, label_transform=None, 
                      normalize=True, flipX=True, flipY=False, 
                      rotate=3, scale_min=0.75, scale_max=1.25, offset=0.1,
                      random_seed = None, numTries=3):

    pin_memory = False
    # if cuda:
    #     pin_memory = True

    startIndex = 0
    allocated = 0.0

    for i, loader in enumerate(loaders):
        dataset = CocoDataset(s3=s3, bucket=bucket, 
                    dataset_desc=loader['dataset'], 
                    class_dictionary=class_dict,
                    image_paths=loader['image_path'],
                    height=height, width=width, 
                    image_transform=image_transform, label_transform=label_transform, 
                    normalize=normalize,  enable_transform=loader['enable_transform'], 
                    flipX=flipX, flipY=flipY, 
                    rotate=rotate, scale_min=scale_min, scale_max=scale_max, offset=offset, numTries=numTries)

        # Creating PT data samplers and loaders:
        loader['batches'] =int(dataset.__len__()/batch_size)
        loader['length'] = loader['batches']*batch_size
        loader['dataset_dfn'] = {}
        collate_fn = functools.partial(collate_fn_replace_corrupted, dataset=dataset, batch_size=batch_size)
        loader['dataloader'] = torch.utils.data.DataLoader(dataset=dataset, 
                                            batch_size=batch_size,
                                            shuffle=shuffle,
                                            drop_last=True,
                                            num_workers=num_workers, 
                                            pin_memory=pin_memory,
                                            timeout=timeout,
                                            collate_fn=collate_fn)         

    return loaders

def main(args):

    s3, creds, s3def = Connect(args.credentails)

    dataset_desc = s3.GetDict(s3def['sets']['dataset']['bucket'],args.dataset_train)
    class_dictionary = s3.GetDict(s3def['sets']['dataset']['bucket'],args.class_dict) 
    imUtil = ImUtil(dataset_desc, class_dictionary)

    parameters = ReadDict(args.test_config)

    loaders_dfn = [{'set':'train', 'dataset': args.dataset_train, 'image_path': args.train_image_path, 'enable_transform':True},
                    {'set':'test', 'dataset':  args.dataset_val, 'image_path': args.val_image_path, 'enable_transform':False}]

    if args.test_iterator:
        store = CocoStore(s3, bucket=s3def['sets']['dataset']['bucket'], 
                        dataset_desc=args.dataset_train, 
                        image_paths=args.train_image_path, 
                        class_dictionary = args.class_dict)

        for i, iman in tqdm(enumerate(store), desc='COCO read image', total=len(store),bar_format='{desc:<8.5}{percentage:3.0f}%|{bar:50}{r_bar}'):
            assert(iman['img'] is not None)
            assert(iman['ann'] is not None)

            if args.num_images > 0 and i >= args.num_images:
                print ('test_iterator complete')
                break

    if args.view_iterator:
        os.makedirs(args.test_path, exist_ok=True)
        
        store = CocoStore(s3, bucket=s3def['sets']['dataset']['bucket'], 
                          dataset_desc=args.dataset_train, 
                          image_paths=args.train_image_path, 
                          class_dictionary=args.class_dict, 
                          imflags=args.imflags, numTries=args.numTries)

        for i, iman in enumerate(store):
            img = store.MergeIman(iman['img'], iman['ann'])
            write_path = '{}cocostoreiterator{:03d}.png'.format(args.test_path, i)
            cv2.imwrite(write_path,img)
            if args.num_images > 0 and i >= args.num_images:
                print ('test_iterator complete')
                break

    if args.test_dataset:
        loaders = CreateCocoLoaders(s3, bucket=s3def['sets']['dataset']['bucket'],
                                    class_dict=args.class_dict, 
                                    batch_size=args.batch_size, 
                                    num_workers=args.num_workers,
                                    cuda = args.cuda,
                                    loaders = loaders_dfn,
                                    height = args.height, width = args.width,
                                    numTries=args.numTries
                                   )

        for loader in tqdm(loaders, desc="Loader"):
            for i, data in tqdm(enumerate(loader['dataloader']), 
                                desc="Batch Reads", 
                                total=loader['batches']):
                inputs, labels, mean, stdev = data
                assert(inputs is not None)
                assert(labels is not None)

            if args.num_images > 0 and i >= args.num_images:
                print ('test_iterator complete')
                break

    if args.view_dataset:

        loaders = CreateCocoLoaders(s3=s3, 
                                    bucket=s3def['sets']['dataset']['bucket'],
                                    class_dict=args.class_dict, 
                                    batch_size=args.batch_size, 
                                    num_workers=args.num_workers, 
                                    cuda = args.cuda,
                                    loaders = loaders_dfn,
                                    height = args.height, width = args.width,
                                    numTries=args.numTries
                                )
        os.makedirs(args.test_path, exist_ok=True)

        for iDataset, loader in enumerate(loaders):
            for i, data  in enumerate(loader['dataloader']):
                images, labels, mean, stdev = data
                images = images.cpu().permute(0, 2, 3, 1).numpy()
                images = np.squeeze(images)
                labels = labels.cpu().numpy()

                for j in  range(args.batch_size):
                    img = imUtil.MergeIman(images[j], labels[j], mean[j].item(), stdev[j].item())
                    write_path = '{}cocostoredataset{}{:03d}{:03d}.png'.format(args.test_path, loader['set'], i,j)
                    cv2.imwrite(write_path,img)
                if args.num_images > 0 and  i > min(np.ceil(args.num_images/args.batch_size), loader['batches']):
                    break
        print ('test_dataset complete')

    print('Test complete')

#objdict = json.load(open('/data/git/mllib/datasets/coco.json'))
#Test(objdict, '/store/Datasets/coco/instances_val2017.json', '/store/Datasets/coco/val2014', 'COCO_val2014_')

def parse_arguments():
    parser = argparse.ArgumentParser(description='Process arguments')

    parser.add_argument('-d', '--debug', action='store_true',help='Wait for debuggee attach')   
    parser.add_argument('-debug_port', type=int, default=3000, help='Debug port')
    parser.add_argument('-debug_address', type=str, default='0.0.0.0', help='Debug port')
    parser.add_argument('-dataset_path', type=str, default='./dataset', help='Local dataset path')
    parser.add_argument('-credentails', type=str, default='creds.yaml', help='Credentials file.')
    parser.add_argument('-dataset_train', type=str, default='data/coco/annotations/instances_train2017.json', help='Coco dataset train instance json file.')
    parser.add_argument('-dataset_val', type=str, default='data/coco/annotations/instances_val2017.json', help='Coco dataset validation instance json file.')
    parser.add_argument('-train_image_path', type=str, default='data/coco/train2017', help='Coco image path for dataset.')
    parser.add_argument('-val_image_path', type=str, default='data/coco/val2017', help='Coco image path for dataset.')
    parser.add_argument('-class_dict', type=str, default='model/segmin/coco.json', help='Model class definition file.')
    parser.add_argument('-num_images', type=int, default=0, help='Number of images to display')
    parser.add_argument('-num_workers', type=int, default=1, help='Data loader workers')
    parser.add_argument('-batch_size', type=int, default=4, help='Dataset batch size')
    parser.add_argument('-i', action='store_true', help='True to test iterator')
    parser.add_argument('-test_iterator', type=bool, default=False, help='True to test iterator')
    parser.add_argument('-ds', action='store_true', help='True to test dataset')
    parser.add_argument('-test_dataset', type=bool, default=True, help='True to test dataset')
    parser.add_argument('-view_iterator', action='store_true', help='True to test iterator')
    parser.add_argument('-view_dataset', action='store_true', help='True to test dataset')
    parser.add_argument('-test_path', type=str, default='./datasets_test/', help='Test path ending in a forward slash')
    parser.add_argument('-test_config', type=str, default='test.yaml', help='Test configuration file')

    parser.add_argument('-height', type=int, default=480, help='Batch image height')
    parser.add_argument('-width', type=int, default=640, help='Batch image width')
    parser.add_argument('-imflags', type=int, default=cv2.IMREAD_COLOR, help='cv2.imdecode flags')
    parser.add_argument('-cuda', type=bool, default=True, help='pytorch CUDA flag') 
    parser.add_argument('-numTries', type=int, default=3, help="Read retries")

    args = parser.parse_args()

    if args.i:
        args.test_iterator = True

    if args.ds:
        args.test_dataset = True

    return args
    
if __name__ == '__main__':
    import argparse
    args = parse_arguments()

    if args.debug:
        print("Wait for debugger attach on {}:{}".format(args.debug_address, args.debug_port))
        import debugpy

        debugpy.listen(address=(args.debug_address, args.debug_port)) # Pause the program until a remote debugger is attached
        debugpy.wait_for_client()  # Pause the program until a remote debugger is attached
        print("Debugger attached")

    main(args)




import sys
import os
from pycocotools import mask
import numpy as np
import cv2
import json
import unittest
from collections import defaultdict
import torch
from torch.utils.data import Dataset
from torchvision import datasets, transforms
from torch.utils.data import DataLoader
from tqdm import tqdm

from pymlutil.s3 import s3store, Connect
from pymlutil.jsonutil import ReadDict
from pymlutil.imutil import ImUtil, ImTransform

from .cocostore import *
from .imstore import *
from .getcoco import *
from .getsceneflow import *

def parse_arguments():
    parser = argparse.ArgumentParser(description='Process arguments')

    parser.add_argument('-d', '--debug', action='store_true',help='Wait for debuggee attach')   
    parser.add_argument('-debug_port', type=int, default=3000, help='Debug port')
    parser.add_argument('-debug_address', type=str, default='0.0.0.0', help='Debug port')
    parser.add_argument('-dataset_path', type=str, default='./dataset', help='Local dataset path')
    parser.add_argument('-credentials', type=str, default='creds.yaml', help='Credentials file.')
    parser.add_argument('-dataset_train', type=str, default='data/coco/annotations/instances_train2017.json', help='Coco dataset train instance json file.')
    parser.add_argument('-dataset_val', type=str, default='data/coco/annotations/instances_val2017.json', help='Coco dataset validation instance json file.')
    parser.add_argument('-train_image_path', type=str, default='data/coco/train2017', help='Coco image path for dataset.')
    parser.add_argument('-val_image_path', type=str, default='data/coco/val2017', help='Coco image path for dataset.')
    parser.add_argument('-class_dict', type=str, default='model/segmin/coco.json', help='Model class definition file.')
    parser.add_argument('-num_images', type=int, default=0, help='Number of images to display.  0 for full dataset')
    parser.add_argument('-num_workers', type=int, default=1, help='Data loader workers')
    parser.add_argument('-batch_size', type=int, default=4, help='Dataset batch size')
    parser.add_argument('-num_tries', type=int, default=3, help='Image read retries')
    

    parser.add_argument('-ci', action='store_true', help='True to test iterator')
    parser.add_argument('-coco_iterator', type=bool, default=False, help='True to test iterator')
    parser.add_argument('-cd', action='store_true', help='True to test dataset')
    parser.add_argument('-coco_dataset', action='store_true', help='True to test dataset')

    parser.add_argument('-coco_dataset_train', type=str, default='data/coco/annotations/instances_train2017.json', help='')
    parser.add_argument('-coco_dataset_val', type=str, default='data/coco/annotations/instances_val2017.json', help='')
    parser.add_argument('-coco_train_image_path', type=str, default='data/coco/train2017', help='')
    parser.add_argument('-coco_val_image_path', type=str, default='data/coco/val2017', help='')
    parser.add_argument('-coco_class_dict', type=str, default='model/segmin/coco.json', help='')

    parser.add_argument('-ii', action='store_true', help='True to test iterator')
    parser.add_argument('-image_iterator', type=bool, default=False, help='True to test iterator')
    parser.add_argument('-id', action='store_true', help='True to test dataset')
    parser.add_argument('-image_dataset', action='store_true', help='True to test dataset')

    parser.add_argument('-image_dataset_desc', type=str, default='data/lit/dataset.yaml', help='')
    parser.add_argument('-image_class_dict', type=str, default='model/crisplit/lit.json', help='')

    parser.add_argument('-test_path', type=str, default='./datasets_test/', help='Test path ending in a forward slash')

    parser.add_argument('-height', type=int, default=640, help='Batch image height')
    parser.add_argument('-width', type=int, default=640, help='Batch image width')
    parser.add_argument('-imflags', type=int, default=cv2.IMREAD_COLOR, help='cv2.imdecode flags')
    parser.add_argument('-cuda', type=bool, default=True)

    parser.add_argument('-getcoco', action='store_true',help='Get coco dataset') 
    parser.add_argument('-cocourl', type=json.loads, default=None, 
                        help='List of coco dataset URLs to load.  If none, coco 2017 datafiles are loaded from https://cocodataset.org/#download')
    parser.add_argument('-cocodatasetname', type=str, default='coco', help='coco dataset name in objet storage')

    parser.add_argument('-getsceneflow', action='store_true',help='Get sceneflow datasets')
    parser.add_argument('-sceneflowurl', type=json.loads, default=None, 
                        help='List of sceneflow dataset URLs to load.  If none, sceneflow_urls are loaded')
    parser.add_argument('-sceneflowdatasetname', type=str, default='sceneflow', help='Sintel dataset name in objet storage')

    args = parser.parse_args()

    if args.ci:
        args.coco_iterator = True

    if args.cd:
        args.coco_dataset = True

    if args.ii:
        args.image_iterator = True

    if args.id:
        args.image_dataset = True
    
    return args


def main(args):

    s3, creds, s3def = Connect(args.credentials)

    if args.getcoco:
        if args.cocourl is not None:
            getcoco(s3, s3def, cocourl=args.cocourl, dataset=args.cocodatasetname)
        else:
            getcoco(s3, s3def, dataset=args.cocodatasetname)

    if args.getsceneflow:
        if args.sceneflowurl is not None:
            getsceneflow(s3, s3def, cocourl=args.sceneflowurl, dataset=args.sceneflowdatasetname)
        else:
            getsceneflow(s3, s3def, dataset=args.sceneflowdatasetname)

        dataset_desc = s3.GetDict(s3def['sets']['dataset']['bucket'],args.dataset_train)
        class_dictionary = s3.GetDict(s3def['sets']['dataset']['bucket'],args.class_dict) 
        imUtil = ImUtil(dataset_desc, class_dictionary)

    if args.coco_iterator:

        store = CocoStore(s3, bucket=s3def['sets']['dataset']['bucket'], 
                        dataset_desc=args.coco_dataset_train, 
                        image_paths=args.coco_train_image_path, 
                        class_dictionary=args.coco_class_dict, numTries=args.num_tries,)

        for i, iman in tqdm(enumerate(store), desc='COCO read image', total=len(store),bar_format='{desc:<8.5}{percentage:3.0f}%|{bar:50}{r_bar}'):
            assert(iman['img'] is not None)
            assert(iman['ann'] is not None)

            if args.num_images > 0 and i >= args.num_images:
                print ('test_iterator complete')
                break

    if args.coco_dataset:
        loaders_dfn = [{'set':'train', 'dataset': args.dataset_train, 'image_path': args.train_image_path, 'enable_transform':True},
                       {'set':'test', 'dataset':  args.dataset_val, 'image_path': args.val_image_path, 'enable_transform':False}]

        loaders = CreateCocoLoaders(s3=s3, 
                                    bucket=s3def['sets']['dataset']['bucket'],
                                    class_dict=args.class_dict, 
                                    batch_size=args.batch_size, 
                                    num_workers=args.num_workers, 
                                    cuda = args.cuda,
                                    loaders = loaders_dfn,
                                    height = args.height, width = args.width,
                                    numTries=args.num_tries
                                )
        os.makedirs(args.test_path, exist_ok=True)

        for loader in tqdm(loaders, desc="Loader"):
            for i, data in tqdm(enumerate(loader['dataloader']), 
                                desc="Batch Reads", 
                                total=loader['batches']):
                inputs, labels, mean, stdev = data
                assert(inputs is not None)
                assert(labels is not None)

            if args.num_images > 0 and i >= args.num_images:
                print ('test_iterator complete')
                break


        for iDataset, loader in enumerate(loaders):
            for i, data  in enumerate(loader['dataloader']):
                images, labels, mean, stdev = data
                images = images.cpu().permute(0, 2, 3, 1).numpy()
                images = np.squeeze(images)
                labels = labels.cpu().numpy()

                for j in  range(args.batch_size):
                    img = imUtil.MergeIman(images[j], labels[j], mean[j].item(), stdev[j].item())
                    write_path = '{}cocostoredataset{}{:03d}{:03d}.png'.format(args.test_path, loader['set'], i,j)
                    cv2.imwrite(write_path,img)
                if args.num_images > 0 and  i > min(np.ceil(args.num_images/args.batch_size), loader['batches']):
                    break
        print ('coco_dataset complete')

    if args.image_iterator:

        store = ImagesStore(s3, bucket=s3def['sets']['dataset']['bucket'], 
                        dataset_desc=args.image_dataset_desc, 
                        class_dictionary=args.image_class_dict, numTries=args.num_tries)
        failures = []
        for i, iman in  tqdm(enumerate(store),
                             desc="ImagesStore iterator reads",
                             total=len(store)):
            #img = store.MergeIman(iman['img'], iman['ann'])
            if iman is None:
                failures.append(i)
                print('torchdatasetutil ImagesStore failed to load image {}: {}'.format(i, store.images[i]))
            elif iman['ann'] is None:
                print('torchdatasetutil ImagesStore failed to load annotation {}: {}'.format(i, store.labels[i]))
                failures.append(i)
            if args.num_images > 0 and i >= args.num_images:
                print ('test_iterator complete')
                break

        print ('{} read failures'.format(len(failures)))
    if args.image_dataset:

        loaders = CreateImageLoaders(s3=s3, 
                                    bucket=s3def['sets']['dataset']['bucket'],
                                    dataset_dfn=args.image_dataset_desc,
                                    class_dict=args.image_class_dict, 
                                    batch_size=args.batch_size, 
                                    num_workers=args.num_workers, 
                                    cuda = args.cuda,
                                    height = args.height, width = args.width,
                                    numTries=args.num_tries,
                                )
        for loader in tqdm(loaders, desc="Loader"):
            for i, data in tqdm(enumerate(loader['dataloader']), 
                                desc="Batch Reads", 
                                total=loader['batches']):
                inputs, labels, mean, stdev = data
                assert(inputs is not None)
                assert(labels is not None)
                assert(inputs.shape[0] == args.batch_size)
                assert(inputs.shape[-1] == args.width)
                assert(inputs.shape[-2] == args.height)

            if args.num_images > 0 and i >= args.num_images:
                print ('test_iterator complete')
                break

    print('torchdatasetutil complete')

#objdict = json.load(open('/data/git/mllib/datasets/coco.json'))
#Test(objdict, '/store/Datasets/coco/instances_val2017.json', '/store/Datasets/coco/val2014', 'COCO_val2014_')


    
if __name__ == '__main__':
    import argparse
    args = parse_arguments()

    if args.debug:
        print("Wait for debugger attach on {}:{}".format(args.debug_address, args.debug_port))
        import debugpy

        debugpy.listen(address=(args.debug_address, args.debug_port)) # Pause the program until a remote debugger is attached
        debugpy.wait_for_client()  # Pause the program until a remote debugger is attached
        print("Debugger attached")

    main(args)





# From torchvision.datasets.cityscapes



import json
import os
import sys
from collections import namedtuple
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import cv2

import torch
from tqdm import tqdm
from torch.utils.data import Dataset
from torchvision import datasets, transforms
from torch.utils.data import DataLoader, WeightedRandomSampler
from torchvision import datasets, transforms
from torchvision.datasets.utils import extract_archive, iterable_to_str, verify_str_arg
from torchvision.datasets.vision import VisionDataset

from pymlutil.s3 import s3store, Connect
from pymlutil.jsonutil import ReadDict
from pymlutil.imutil import ImUtil, ImTransform

sys.path.insert(0, os.path.abspath('')) # Test files from current path rather than installed module
from torchdatasetutil.getcityscapes import getcityscapes

class CityscapesDataset(VisionDataset):
    """`Cityscapes <http://www.cityscapes-dataset.com/>`_ Dataset.

    Args:
        root (string): Root directory of dataset where directory ``leftImg8bit``
            and ``gtFine`` or ``gtCoarse`` are located.
        split (string, optional): The image split to use, ``train``, ``test`` or ``val`` if mode="fine"
            otherwise ``train``, ``train_extra`` or ``val``
        mode (string, optional): The quality mode to use, ``fine`` or ``coarse``
        target_type (string or list, optional): Type of target to use, ``instance``, ``semantic``, ``polygon``
            or ``color``. Can also be a list to output a tuple with all specified target types.
        transform (callable, optional): A function/transform that takes in a PIL image
            and returns a transformed version. E.g, ``transforms.RandomCrop``
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it.
        transforms (callable, optional): A function/transform that takes input sample and its target as entry
            and returns a transformed version.

    Examples:

        Get semantic segmentation target

        .. code-block:: python

            dataset = Cityscapes('./data/cityscapes', split='train', mode='fine',
                                 target_type='semantic')

            img, smnt = dataset[0]

        Get multiple targets

        .. code-block:: python

            dataset = Cityscapes('./data/cityscapes', split='train', mode='fine',
                                 target_type=['instance', 'color', 'polygon'])

            img, (inst, col, poly) = dataset[0]

        Validate on the "coarse" set

        .. code-block:: python

            dataset = Cityscapes('./data/cityscapes', split='val', mode='coarse',
                                 target_type='semantic')

            img, smnt = dataset[0]
    """

    # Based on https://github.com/mcordts/cityscapesScripts
    CityscapesClass = namedtuple(
        "CityscapesClass",
        ["name", "id", "train_id", "category", "category_id", "has_instances", "ignore_in_eval", "color"],
    )

    classes = [
        CityscapesClass("unlabeled", 0, 255, "void", 0, False, True, (0, 0, 0)),
        CityscapesClass("ego vehicle", 1, 255, "void", 0, False, True, (0, 0, 0)),
        CityscapesClass("rectification border", 2, 255, "void", 0, False, True, (0, 0, 0)),
        CityscapesClass("out of roi", 3, 255, "void", 0, False, True, (0, 0, 0)),
        CityscapesClass("static", 4, 255, "void", 0, False, True, (0, 0, 0)),
        CityscapesClass("dynamic", 5, 255, "void", 0, False, True, (111, 74, 0)),
        CityscapesClass("ground", 6, 255, "void", 0, False, True, (81, 0, 81)),
        CityscapesClass("road", 7, 0, "flat", 1, False, False, (128, 64, 128)),
        CityscapesClass("sidewalk", 8, 1, "flat", 1, False, False, (244, 35, 232)),
        CityscapesClass("parking", 9, 255, "flat", 1, False, True, (250, 170, 160)),
        CityscapesClass("rail track", 10, 255, "flat", 1, False, True, (230, 150, 140)),
        CityscapesClass("building", 11, 2, "construction", 2, False, False, (70, 70, 70)),
        CityscapesClass("wall", 12, 3, "construction", 2, False, False, (102, 102, 156)),
        CityscapesClass("fence", 13, 4, "construction", 2, False, False, (190, 153, 153)),
        CityscapesClass("guard rail", 14, 255, "construction", 2, False, True, (180, 165, 180)),
        CityscapesClass("bridge", 15, 255, "construction", 2, False, True, (150, 100, 100)),
        CityscapesClass("tunnel", 16, 255, "construction", 2, False, True, (150, 120, 90)),
        CityscapesClass("pole", 17, 5, "object", 3, False, False, (153, 153, 153)),
        CityscapesClass("polegroup", 18, 255, "object", 3, False, True, (153, 153, 153)),
        CityscapesClass("traffic light", 19, 6, "object", 3, False, False, (250, 170, 30)),
        CityscapesClass("traffic sign", 20, 7, "object", 3, False, False, (220, 220, 0)),
        CityscapesClass("vegetation", 21, 8, "nature", 4, False, False, (107, 142, 35)),
        CityscapesClass("terrain", 22, 9, "nature", 4, False, False, (152, 251, 152)),
        CityscapesClass("sky", 23, 10, "sky", 5, False, False, (70, 130, 180)),
        CityscapesClass("person", 24, 11, "human", 6, True, False, (220, 20, 60)),
        CityscapesClass("rider", 25, 12, "human", 6, True, False, (255, 0, 0)),
        CityscapesClass("car", 26, 13, "vehicle", 7, True, False, (0, 0, 142)),
        CityscapesClass("truck", 27, 14, "vehicle", 7, True, False, (0, 0, 70)),
        CityscapesClass("bus", 28, 15, "vehicle", 7, True, False, (0, 60, 100)),
        CityscapesClass("caravan", 29, 255, "vehicle", 7, True, True, (0, 0, 90)),
        CityscapesClass("trailer", 30, 255, "vehicle", 7, True, True, (0, 0, 110)),
        CityscapesClass("train", 31, 16, "vehicle", 7, True, False, (0, 80, 100)),
        CityscapesClass("motorcycle", 32, 17, "vehicle", 7, True, False, (0, 0, 230)),
        CityscapesClass("bicycle", 33, 18, "vehicle", 7, True, False, (119, 11, 32)),
        #CityscapesClass("license plate", -1, -1, "vehicle", 7, False, True, (0, 0, 142)),
    ]

    def __init__(
        self,
        root: str,
        class_dictionary: dict,
        split: str = "train",
        mode: str = "fine",
        target_type: Union[List[str], str] = "instance",
        transforms: Optional[Callable] = None,
        imflags = cv2.IMREAD_COLOR,
        anflags = cv2.IMREAD_GRAYSCALE,
        flipX=True, 
        flipY=False, 
        rotate=3, 
        scale_min=0.75, 
        scale_max=1.25, 
        offset=0.1,
    ) -> None:
        super().__init__(root)
        self.class_dictionary = class_dictionary
        self.mode = "gtFine" if mode == "fine" else "gtCoarse"
        self.images_dir = os.path.join(self.root, "leftImg8bit", split)
        self.targets_dir = os.path.join(self.root, self.mode, split)
        self.target_type = target_type
        self.split = split
        self.images = []
        self.targets = []
        self.imflags = imflags
        self.anflags = anflags
        self.transforms = transforms

        verify_str_arg(mode, "mode", ("fine", "coarse"))
        if mode == "fine":
            valid_modes = ("train", "test", "val")
        else:
            valid_modes = ("train", "train_extra", "val")
        msg = "Unknown value '{}' for argument split if mode is '{}'. Valid values are {{{}}}."
        msg = msg.format(split, mode, iterable_to_str(valid_modes))
        verify_str_arg(split, "split", valid_modes, msg)

        if not isinstance(target_type, list):
            self.target_type = [target_type]
        [
            verify_str_arg(value, "target_type", ("instance", "semantic", "polygon", "color"))
            for value in self.target_type
        ]

        if not os.path.isdir(self.images_dir) or not os.path.isdir(self.targets_dir):

            if split == "train_extra":
                image_dir_zip = os.path.join(self.root, "leftImg8bit_trainextra.zip")
            else:
                image_dir_zip = os.path.join(self.root, "leftImg8bit_trainvaltest.zip")

            if self.mode == "gtFine":
                target_dir_zip = os.path.join(self.root, f"{self.mode}_trainvaltest.zip")
            elif self.mode == "gtCoarse":
                target_dir_zip = os.path.join(self.root, f"{self.mode}.zip")

            if os.path.isfile(image_dir_zip) and os.path.isfile(target_dir_zip):
                extract_archive(from_path=image_dir_zip, to_path=self.root)
                extract_archive(from_path=target_dir_zip, to_path=self.root)
            else:
                raise RuntimeError(
                    "Dataset not found or incomplete. Please make sure all required folders for the"
                    ' specified "split" and "mode" are inside the "root" directory'
                )

        for city in os.listdir(self.images_dir):
            img_dir = os.path.join(self.images_dir, city)
            target_dir = os.path.join(self.targets_dir, city)
            for file_name in os.listdir(img_dir):
                target_types = []
                for t in self.target_type:
                    target_name = "{}_{}".format(
                        file_name.split("_leftImg8bit")[0], self._get_target_suffix(self.mode, t)
                    )
                    target_types.append(os.path.join(target_dir, target_name))

                self.images.append(os.path.join(img_dir, file_name))
                self.targets.append(target_types)

    def classes(self, anns):
        class_vector = np.zeros(self.class_dictionary ['classes'], dtype=np.float32)

        for ann in anns:
            obj = self.catToObj[ann['category_id']]
            if obj['trainId'] < self.class_dictionary ["classes"]:
                class_vector[obj['trainId']] = 1.0

        return class_vector

    def ConvertLabels(self, ann):
        trainAnn = np.zeros_like(ann)
        for obj in self.class_dictionary ['objects']: # Load RGB colors as BGR
            if not (obj['id'] == obj['trainId'] and obj['id'] == 0):
                trainAnn[ann==obj['id']] = obj['trainId']
        return trainAnn

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        """
        Args:
            index (int): Index
        Returns:
            tuple: (image, target) where target is a tuple of all target types if target_type is a list with more
            than one item. Otherwise target is a json object if target_type="polygon", else the image segmentation.
        """

        image = cv2.imread(self.images[index], self.imflags)

        targets: Any = []
        for i, t in enumerate(self.target_type):
            if t == "polygon":
                target = self._load_json(self.targets[index][i])
            else:
                target = cv2.imread(self.targets[index][i], self.anflags)

            if t == "semantic":
                target = self.ConvertLabels(target)

            targets.append(target)

        target = tuple(targets) if len(targets) > 1 else targets[0]

        if self.transforms is not None:
            image, target, mean, std = self.transforms(image, target)
        else:
            mean = None
            std = None

        image = torch.from_numpy(image).permute(2, 0, 1)
        target = torch.from_numpy(target)

        assert(image.shape[-1] == self.transforms.width)
        assert(image.shape[-2] == self.transforms.height)

        return image, target, mean, std

    def __len__(self) -> int:
        return len(self.images)

    def extra_repr(self) -> str:
        lines = ["Split: {split}", "Mode: {mode}", "Type: {target_type}"]
        return "\n".join(lines).format(**self.__dict__)

    def _load_json(self, path: str) -> Dict[str, Any]:
        with open(path) as file:
            data = json.load(file)
        return data

    def _get_target_suffix(self, mode: str, target_type: str) -> str:
        if target_type == "instance":
            return f"{mode}_instanceIds.png"
        elif target_type == "semantic":
            return f"{mode}_labelIds.png"
        elif target_type == "color":
            return f"{mode}_color.png"
        else:
            return f"{mode}_polygons.json"

def CreateCityscapesLoaders(s3, s3def, src, dest, class_dictionary, bucket = None, width=256, height=256, batch_size = 2, shuffle=True, 
                      num_workers=0, cuda = True, timeout=0, loaders = None, 
                      transforms=True, 
                      normalize=True, 
                      random_seed = None, numTries=3,
                      flipX=True, 
                      flipY=False, 
                      rotate=3, 
                      scale_min=0.75, 
                      scale_max=1.25, 
                      offset=0.1,
                      astype='float32',
                      borderType=cv2.BORDER_CONSTANT,
                      borderValue=0,
                      train_sampler_weights=None):
    if not bucket:
        bucket = s3def['sets']['dataset']['bucket']

    if not os.path.exists(dest):
        s3.GetDir(bucket, src, dest)

    dest = os.path.join(dest, '') #Ensure there is a trailing slash

    pin_memory = False
    # if cuda:
    #     pin_memory = True

    # Load dataset
    if loaders is None:

        default_loaders = [{'set':'train', 'dataset': dest, 'enable_transform':True, 'mode':'fine', 'target_type':['semantic'], 'class_dictionary':class_dictionary } ,
                        {'set':'val', 'dataset': dest, 'enable_transform':False, 'mode':'fine', 'target_type':['semantic'], 'class_dictionary':class_dictionary}]

        loaders = default_loaders

    startIndex = 0
    allocated = 0.0

    for i, loader in enumerate(loaders):

        if train_sampler_weights is not None and loader['set'] == 'train':
            sampler=WeightedRandomSampler(weights=train_sampler_weights, num_samples=len(train_sampler_weights), replacement=True)
            shuffle=False
            print('Weighted Random Sampler is initiated!')
        else:
            sampler=None


        if transforms and loader['set'] == 'train':
            transform = ImTransform(height=height, width=width, 
                        normalize=normalize, 
                        enable_transform=loader['enable_transform'],
                        flipX=flipX, 
                        flipY=flipY, 
                        rotate=rotate, 
                        scale_min=scale_min, 
                        scale_max=scale_max, 
                        offset=offset, 
                        astype=astype,
                        borderType=borderType,
                        borderValue=borderValue,
                        )
        elif transforms:
            transform = ImTransform(height=height, width=width, 
                        normalize=normalize, 
                        astype=astype,
                        enable_transform=False, 
                        borderType=borderType,
                        borderValue=borderValue,
                        )
            


        dataset = CityscapesDataset(root=loader['dataset'],
                                    class_dictionary=loader['class_dictionary'],
                                    split=loader['set'], 
                                    mode=loader['mode'], 
                                    target_type=loader['target_type'], 
                                    transforms=transform
                                   )
        


        loader['dataloader'] = torch.utils.data.DataLoader(dataset=dataset,
                                                batch_size=batch_size,
                                                shuffle=shuffle,
                                                num_workers=num_workers,
                                                pin_memory=pin_memory,
                                                sampler=sampler)

        # Creating PT data samplers and loaders:
        loader['batches'] =len(loader['dataloader'])
        loader['length'] = loader['batches']*batch_size
        loader['width']=width
        loader['height']=height
        loader['in_channels']=loader['class_dictionary']['input_channels']
        loader['num_classes']=loader['class_dictionary']['classes']
        loader['classes']=loader['class_dictionary']['objects']
        loader['dataset_dfn'] = {}

    return loaders


import sys
import argparse
import json
import os
import subprocess
import shutil
from os import fspath
import tempfile
from pymlutil.s3 import s3store, Connect
from zipfile import ZipFile
from tqdm import tqdm


cocourl=["http://images.cocodataset.org/zips/train2017.zip",
         "http://images.cocodataset.org/zips/val2017.zip",
         "http://images.cocodataset.org/zips/test2017.zip",
         "http://images.cocodataset.org/zips/unlabeled2017.zip",
         "http://images.cocodataset.org/annotations/annotations_trainval2017.zip",
         "http://images.cocodataset.org/annotations/stuff_annotations_trainval2017.zip",
         "http://images.cocodataset.org/annotations/panoptic_annotations_trainval2017.zip",
         "http://images.cocodataset.org/annotations/image_info_test2017.zip",
         "http://images.cocodataset.org/annotations/image_info_unlabeled2017.zip",
        ]

def getcoco(s3, s3def, cocourl=cocourl, dataset='coco'):
    saved_name = '{}/{}'.format(s3def['sets']['dataset']['prefix'] , dataset)
    for url in tqdm(cocourl, bar_format='{desc:<8.5}{percentage:3.0f}%|{bar:50}{r_bar}'):
        with tempfile.TemporaryDirectory() as tmpdir:
            outpath = '{}/{}'.format(tmpdir,os.path.basename(url))
            if os.path.isfile(outpath):
                print('{} exists.  Skipping'.format(outpath))
            else:
                sysmsg = 'wget -O {} {} '.format(outpath, url)
                print(sysmsg)
                os.system(sysmsg)

            dest = '{}/{}'.format(tmpdir,dataset)
            with ZipFile(outpath,"r") as zip_ref:
                for file in tqdm(iterable=zip_ref.namelist(), 
                    bar_format='{desc:<8.5}{percentage:3.0f}%|{bar:50}{r_bar}', 
                    total=len(zip_ref.namelist())):
                    zip_ref.extract(member=file, path=fspath(dest))

            os.remove(outpath) # Remove zip

            print('Save {} to {}/{}'.format(dest, s3def['sets']['dataset']['bucket'], saved_name))
            s3.PutDir(s3def['sets']['dataset']['bucket'], dest, saved_name)

    url = s3.GetUrl(s3def['sets']['dataset']['bucket'], saved_name)
    print("Complete. Results saved to {}".format(url))

def parse_arguments():
    parser = argparse.ArgumentParser(description='Process arguments')

    parser.add_argument('-d', '--debug', action='store_true',help='Wait for debuggee attach')   
    parser.add_argument('-debug_port', type=int, default=3000, help='Debug port')
    parser.add_argument('-debug_address', type=str, default='0.0.0.0', help='Debug port')
    parser.add_argument('-dataset_path', type=str, default='./dataset', help='Local dataset path')
    parser.add_argument('-credentails', type=str, default='creds.yaml', help='Credentials file.')
    parser.add_argument('-min', action='store_true',help='Minimum test')
    parser.add_argument('-cocourl', type=json.loads, default=cocourl, 
                        help='List of coco dataset URLs to load.  If none, coco 2017 datafiles are loaded from https://cocodataset.org/#download')


    args = parser.parse_args()
    return args

def main(args):

    s3, _, s3def = Connect(args.credentails)

    if args.min:
        args.cocourl = [args.cocourl[-2], args.cocourl[-1]]

    getcoco(s3, s3def, cocourl=args.cocourl, dataset='coco_test')

    print('{} {} complete'.format(__file__, __name__))

if __name__ == '__main__':
    import argparse
    args = parse_arguments()

    if args.debug:
        print("Wait for debugger attach on {}:{}".format(args.debug_address, args.debug_port))
        import debugpy

        debugpy.listen(address=(args.debug_address, args.debug_port)) # Pause the program until a remote debugger is attached
        debugpy.wait_for_client()  # Pause the program until a remote debugger is attached
        print("Debugger attached")

    main(args)


import sys
import os
from pycocotools import mask
import numpy as np
import cv2
import json
from collections import defaultdict
import unittest
import torch
from tqdm import tqdm
from torch.utils.data import Dataset
from torchvision import datasets, transforms
from torch.utils.data import DataLoader

from pymlutil.s3 import s3store, Connect
from pymlutil.jsonutil import ReadDict
from pymlutil.imutil import ImUtil, ImTransform

sys.path.insert(0, os.path.abspath('')) # Test files from current path rather than installed module
from torchdatasetutil.imstore import ImagesStore, CreateImageLoaders

test_config = 'test.yaml'
class Test(unittest.TestCase):      

    def test_iterator(self):
        parameters = ReadDict(test_config)

        if 'images' not in parameters:
            raise ValueError('images not in {}'.format(test_config))

        s3, creds, s3def = Connect(parameters['images']['credentials'])
                
        store = ImagesStore(s3, bucket=s3def['sets']['dataset']['bucket'], 
                        dataset_desc=parameters['images']['dataset'], 
                        class_dictionary=parameters['images']['class_dict'])

        for i, iman in  tqdm(enumerate(store),
                             desc="ImagesStore iterator reads",
                             total=len(store)):
            #img = store.MergeIman(iman['img'], iman['ann'])
            assert(iman['img'] is not None)
            assert(iman['ann'] is not None)
            if 'test_images' in parameters['images'] and parameters['images']['test_images'] is not None  and i >= parameters['images']['test_images']:
                break
    
    def test_CreateImageLoaders(self):
        parameters = ReadDict(test_config)

        if 'images' not in parameters:
            raise ValueError('images not in {}'.format(test_config))

        s3, creds, s3def = Connect(parameters['images']['credentials'])

        dataset_desc = s3.GetDict(s3def['sets']['dataset']['bucket'],parameters['images']['dataset'])
        class_dictionary = s3.GetDict(s3def['sets']['dataset']['bucket'],parameters['images']['class_dict']) 
        imUtil = ImUtil(dataset_desc, class_dictionary)

        loaders = CreateImageLoaders(s3, bucket=s3def['sets']['dataset']['bucket'],
                                     dataset_dfn=parameters['images']['dataset'],
                                     batch_size=parameters['images']['batch_size'], 
                                     height=parameters['images']['height'], 
                                     width=parameters['images']['width'],
                                     class_dict=parameters['images']['class_dict'])

        for loader in tqdm(loaders, desc="CreateImageLoaders"):
            for i, data in tqdm(enumerate(loader['dataloader']), 
                                desc="Batch Reads", 
                                total=loader['batches']):
                inputs, labels, mean, stdev = data
                assert(inputs.size(0)==parameters['images']['batch_size'])
                assert(inputs.size(-1)==parameters['images']['width'])
                assert(inputs.size(-2)==parameters['images']['height'])
                assert(labels.size(0)==parameters['images']['batch_size'])
                assert(labels.size(-1)==parameters['images']['width'])
                assert(labels.size(-2)==parameters['images']['height'])

                #images = inputs.cpu().permute(0, 3, 1, 2).numpy()
                #labels = np.around(labels.cpu().numpy()).astype('uint8')
                #mean = mean.cpu().numpy()
                #stdev = stdev.cpu().numpy()

                #for j, image in enumerate(images):
                #    img = imUtil.MergeIman(images[j], labels[j], mean[j], stdev[j])
                #    is_success, buffer = cv2.imencode(".png", img)
                #    if not is_success:
                #        raise ValueError('test_imstore test_CreateImageLoaders cv2.imencode failure batch {} image {}'.format(i, j))
                if 'test_images' in parameters['images'] and parameters['images']['test_images'] is not None and i >= parameters['images']['test_images']:
                    break

if __name__ == '__main__':
    unittest.main()


import sys, os
sys.path.insert(0, os.path.abspath('.'))
from pymlutil.jsonutil import ReadDict

__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
config = ReadDict('{}/build.yaml'.format(__location__))

def VersionString(config):
    version_str = '{}.{}.{}'.format(config['version']['major'], config['version']['minor'], config['version']['patch'])
    if config['version']['label'] is not None and len(config['version']['label']) > 0:
        version_str += '-{}'.format(config['version']['label'])
    return version_str

def version():
    return VersionString(config)

__version__ = version()


import sys
import os
from pycocotools import mask
import numpy as np
import cv2
import json
from collections import defaultdict
import unittest
import torch
from tqdm import tqdm
from torch.utils.data import Dataset
from torchvision import datasets, transforms
from torch.utils.data import DataLoader

from pymlutil.s3 import s3store, Connect
from pymlutil.jsonutil import ReadDict
from pymlutil.imutil import ImUtil, ImTransform

sys.path.insert(0, os.path.abspath('')) # Test files from current path rather than installed module
from torchdatasetutil.imagenetstore import CreateImagenetLoaders

test_config = 'test.yaml'
class Test(unittest.TestCase):      
   
    def test_CreateImagenetLoaders(self):
        parameters = ReadDict(test_config)

        if 'imagenet' not in parameters:
            raise ValueError('images not in {}'.format(test_config))

        s3, creds, s3def = Connect(parameters['imagenet']['credentials'])

        loaders = CreateImagenetLoaders(s3, s3def, 
                                parameters['imagenet']['obj_src'], 
                                parameters['imagenet']['destination'], 
                                augment=parameters['imagenet']['augment'], 
                                normalize=parameters['imagenet']['normalize'], 
                                resize_width=parameters['imagenet']['resize_width'], 
                                resize_height=parameters['imagenet']['resize_height'],
                                crop_width=parameters['imagenet']['crop_width'], 
                                crop_height=parameters['imagenet']['crop_height'], 
                                batch_size=parameters['imagenet']['batch_size'], 
                                num_workers=parameters['imagenet']['num_workers'],
                                flipX=parameters['imagenet']['flipX'], 
                                flipY=parameters['imagenet']['flipY'], 
                                rotate=parameters['imagenet']['rotate'], 
                                scale_min=parameters['imagenet']['scale_min'], 
                                scale_max=parameters['imagenet']['scale_max'], 
                                offset=parameters['imagenet']['offset'], 
                                augment_noise=parameters['imagenet']['augment_noise'])

        parameters['imagenet']['test_path']=os.path.join(parameters['imagenet']['test_path'], '') # Add trailing slash if not present
        os.makedirs(parameters['imagenet']['test_path'], exist_ok=True)

        for loader in tqdm(loaders, desc="Loader"):
            for i, data in tqdm(enumerate(loader['dataloader']), 
                                desc="Batch Reads", 
                                total=loader['batches']):
                sample, target = data
                assert(sample is not None)
                assert(sample.shape[0] == parameters['imagenet']['batch_size'])
                if loader['height'] is not None:
                    assert(sample.shape[2] == loader['height'])
                if loader['width'] is not None:
                    assert(sample.shape[3] == loader['width'])
                    assert(target is not None)

                if parameters['imagenet']['save_image']:
                    sample =  sample.permute(0, 2, 3, 1) # Change to batch, height, width, channel for rendering
                    sample_max = sample.max()
                    sample_min = sample.min()
                    if sample_max > sample_min:
                        for j, image in enumerate(sample):
                            image = 255*(image - sample_min)/(sample_max-sample_min) # Convert to RGB color rane
                            image = image.cpu().numpy().astype(np.uint8)
                            image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
                            write_path = '{}{}{:03d}{:03d}.png'.format(parameters['imagenet']['test_path'], loader['set'], i,j)            
                            cv2.imwrite(write_path,image)

                if parameters['imagenet']['test_images'] is not None and parameters['imagenet']['test_images'] > 0 and (i+1)*parameters['imagenet']['batch_size'] >= parameters['imagenet']['test_images']:
                    print ('test_iterator complete')
                    break

if __name__ == '__main__':
    unittest.main()